//******************************************************************************
//Name          : Regression.cs
//Description   : Implementation to experience the beauty and the pitfalls of Regression
//Author        : Dr. Oliver Kluge, 90411 Nuernberg, Germany
//Copyright(C)  : Oliver Kluge 2014 
//------------------------------------------------------------------------------
//******************************************************************************

using System;
using Display;


class LinearEquation
{
    public double[,] A;		// Koeffizientenmatrix
    public double[]  b;		// Absolutglieder
    public double[]  x;		// Lsungsvektor

    public int dimension;

    // c'tor
    public LinearEquation(int dim)
    {
        dimension = dim;

        A = new double[dimension, dimension];
        b = new double[dimension];
        x = new double[dimension];
    }

    public void FindRefernceRow(int row)
    {
        int column  = row;
        int refrow  = row;
        double refelement = A[row, row];
        double temp       = 0;

        // Alle folgenden Elemente dieser Spalte
        // nach grtem Element durchsuchen
        for (int k=row+1; k<dimension; k++)
        {
            if (Math.Abs(A[k, column]) > Math.Abs(refelement))
            {
                refelement = A[k, column];
                refrow = k;
            }
        }

        // Neue Referenzzeile gefunden?
        // Falls ja, dann Zeilen tauschen
        if (refrow > row)
        {
            // Erst die Koeffizientenmatrix
            for (int k=0; k<dimension; k++)
            {
                temp = A[refrow, k];
                A[refrow, k] = A[row, k];
                A[row, k] = temp;
            }

            // Dann die Absolutglieder
            temp = b[refrow];
            b[refrow] = b[row];
            b[row] = temp;
        }
    }

    public void GaussElimination()
    {
        double norm  = 0;
		
        // Durchlufe um Matrix auf Dreiecksform zu bringen
        for (int i=0; i<dimension; i++)
        {
            // Normierung der Zeilen
            // beginnen mit der (i)ten Zeile
            for (int j=i; j<dimension; j++)
            {
                // Zeile mit grtem Pivot finden
                FindRefernceRow(j);

                // Pivotelement als Norm
                norm = A[j,i];
                if (norm == 0)
                {
	                // Nur zur Sicherheit :-)
	                norm   = 1.0E-20;
                }

                // ber alle verbleibenden Elemente der Zeile
                for (int k=i; k<dimension; k++)
                {
	                A[j,k] /= norm;
                }
                b[j] /= norm;
            }

            // Zeilenweise Reduktion durch Subtraktion
            // beginnend mit der (i+1)ten Zeile
            for (int j=i+1; j<dimension; j++)
            {
                // ber alle verbleibenden Elemente der Zeile
                for (int k=i; k<dimension; k++)
                {
                    A[j,k] -= A[i,k];
                }
                b[j] -= b[i];
            }
        }

        // Einsetzen in Rckwrtsrichtung
        // d.h. zeilenweise, hinten beginnend
        for (int j=dimension-1; j>=0; j--)
        {
            // Ausser Pivotelement alles durch Subtraktion
            // auf die rechte Seite bringen
            for (int k=dimension-1; k>j; k--)
            {
	            b[j] -= x[k]*A[j,k];
            }
            // Unbekannte entspricht dann dem Absolutglied
            x[j] = b[j];
            }
    }
}


class DataModelling
{
    const int samples = 16;

    public double[] y  = new double[samples];   // idealer Signalverlauf
    public double[] yi = new double[samples];   // gemessenes Signal
    public double[] ym = new double[samples];   // modelliertes Signal
    // fr Transformation des Arguments
    public double[] x  = new double[samples];
    
    // Regressionskoeffizienten
    public double a0;
    public double a1;
    public double a2;
    public double a3;
	
    // verbleibende Fehlerquadratsumme der Modellierung
    public double errquadsum;

    // c'tor
    public DataModelling()
    {
        a0 = 0;
        a1 = 0;
        a2 = 0;
        a3 = 0;
        errquadsum = 0;
    }

    public void CreateSignal(int select)
    {
        MyRandom.Random noise = new MyRandom.Random(0.71036, 0.23906);

        // lineare Datenpunkte
        if (select == 1)
        {
            for (int k=0; k<samples; k++)
            {
                // idealer Signalverlauf
                y[k] = 3*k + 10;
                // "gemessener" Signalverlauf
                yi[k]  = y[k] + noise.Normal(0, 3);
            }
        }

        // polygonale (2. Ordnung) Datenpunkte
        if (select == 2)
        {
            for (int k=0; k<samples; k++)
            {
                // idealer Signalverlauf
                y[k] = -0.5*k*k + 7*k + 10.0;
                // "gemessener" Signalverlauf
                yi[k]  = y[k] + noise.Normal(0, 2);
            }
        }

        // polygonale (3. Ordnung) Datenpunkte
        if (select == 3)
        {
            for (int k=0; k<samples; k++)
            {
		        // idealer Signalverlauf
		        y[k] = -0.4*k*k*k + 7*k*k - 20*k + 82.0;
		        // "gemessener" Signalverlauf
                yi[k]  = y[k] + noise.Normal(0, 4);
            }
        }

        // exponentielle Datenpunkte mit a>0
        if (select == 4)
        {
            for (int k=0; k<samples; k++)
            {
		        // idealer Signalverlauf
		        y[k] = 2*Math.Exp(0.3*k);
		        // "gemessener" Signalverlauf
                yi[k]  = y[k] + noise.Normal(0, 2);
            }
        }

        // exponentielle Datenpunkte mit a<0
        if (select == 5)
        {
            for (int k=0; k<samples; k++)
            {
		        // idealer Signalverlauf
		        y[k] = 10*Math.Exp(-0.1*k);
		        // "gemessener" Signalverlauf
                yi[k]  = y[k] + noise.Normal(0, 2);
            }
        }

        // potenzfrmige Datenpunkte mit a>1
        if (select == 6)
        {
            for (int k=0; k<samples; k++)
            {
		        // idealer Signalverlauf
		        y[k] = 3*Math.Pow((double)k, 2.0);
		        // "gemessener" Signalverlauf
                yi[k]  = y[k] + noise.Normal(0, 2);
            }
        }

        // potenzfrmige Datenpunkte mit a<1
        if (select == 7)
        {
            for (int k=0; k<samples; k++)
            {
		        // idealer Signalverlauf
		        y[k] = 5*Math.Pow((double)k, 0.5);
		        // "gemessener" Signalverlauf
                yi[k]  = y[k] + noise.Normal(0, 2);
            }
        }

        // store argument data
        for (int k = 0; k < samples; k++)
        {
            x[k] = k;
        }
    }

    public void LinearRegression()
    {
        double sumx  = 0;
        double sumy  = 0;
        double sumxx = 0;
        double sumxy = 0;

        for (int k=0; k<samples; k++)
        {
            sumx  += x[k];
            sumy  += yi[k];
            sumxx += x[k]*x[k];
            sumxy += x[k]*yi[k];
        }

        // Geradensteigung
        a1 = (samples*sumxy-sumx*sumy)/
             (samples*sumxx-sumx*sumx);
        // Achsenabschnitt
        a0 = (sumy-a1*sumx)/samples;

        // Modelldaten und verbleibende Fehlerquadratsumme berechnen
        for (int k=0; k<samples; k++)
        {
            ym[k] = a1*k + a0;
            errquadsum = (yi[k]-ym[k])*(yi[k]-ym[k]);
        }
    }

    public void Polynom2Regression()
    {
        double sumx    = 0;
        double sumxx   = 0;
        double sumxxx  = 0;
        double sumxxxx = 0;
        double sumy    = 0;
        double sumyx   = 0;
        double sumyxx  = 0;
        double n       = (double)samples;

        for (int k=0; k<samples; k++)
        {
            sumx    += k;
            sumxx   += k*k;
            sumxxx  += k*k*k;
            sumxxxx += k*k*k*k;
            sumy    += yi[k];
            sumyx   += yi[k]*k;
            sumyxx  += yi[k]*k*k;
        }

        a2 = ( (sumyx/sumx-sumy/n)*(sumxxx/sumxx-sumx/n) - 
               (sumyxx/sumxx-sumy/n)*(sumxx/sumx-sumx/n)  )
             /
             ( (sumxxx/sumx-sumxx/n)*(sumxxx/sumxx-sumx/n) -
               (sumxxxx/sumxx-sumxx/n)*(sumxx/sumx-sumx/n)  );

        a1 = ( (sumyxx/sumxx-sumy/n) - a2*(sumxxxx/sumxx-sumxx/n) )
             /
             (sumxxx/sumxx-sumx/n);

        a0 = (sumy - a2*sumxx - a1*sumx) / n;

        // Modelldaten und verbleibende Fehlerquadratsumme berechnen
        for (int k=0; k<samples; k++)
        {
            ym[k] = a2*k*k + a1*k + a0;
            errquadsum = (yi[k]-ym[k])*(yi[k]-ym[k]);
        }
    }

    public void Polynom3Regression()
    {
        double sumx       = 0;
        double sumxx      = 0;
        double sumxxx     = 0;
        double sumxxxx    = 0;
        double sumxxxxx   = 0;
        double sumxxxxxx  = 0;
        double sumy       = 0;
        double sumyx      = 0;
        double sumyxx     = 0;
        double sumyxxx    = 0;
        double n          = (double)samples;
        LinearEquation linear = new LinearEquation(4);

        // Berechnung der Summen
        for (int k=0; k<samples; k++)
        {
            sumx      += k;
            sumxx     += k*k;
            sumxxx    += k*k*k;
            sumxxxx   += k*k*k*k;
            sumxxxxx  += k*k*k*k*k;
            sumxxxxxx += k*k*k*k*k*k;
            sumy      += yi[k];
            sumyx     += yi[k]*k;
            sumyxx    += yi[k]*k*k;
            sumyxxx   += yi[k]*k*k*k;
        }

        // Belegung derKoeffizientenmatrix
        linear.A[0,0] = n;
        linear.A[0,1] = sumx;
        linear.A[0,2] = sumxx;
        linear.A[0,3] = sumxxx;

        linear.A[1,0] = sumx;
        linear.A[1,1] = sumxx;
        linear.A[1,2] = sumxxx;
        linear.A[1,3] = sumxxxx;

        linear.A[2,0] = sumxx;
        linear.A[2,1] = sumxxx;
        linear.A[2,2] = sumxxxx;
        linear.A[2,3] = sumxxxxx;

        linear.A[3,0] = sumxxx;
        linear.A[3,1] = sumxxxx;
        linear.A[3,2] = sumxxxxx;
        linear.A[3,3] = sumxxxxxx;

        // Belegung der Absolutglieder
        linear.b[0] = sumy;
        linear.b[1] = sumyx;
        linear.b[2] = sumyxx;
        linear.b[3] = sumyxxx;

        linear.GaussElimination();

        a0 = linear.x[0];
        a1 = linear.x[1];
        a2 = linear.x[2];
        a3 = linear.x[3];

        // Modelldaten und verbleibende Fehlerquadratsumme berechnen
        for (int k=0; k<samples; k++)
        {
            ym[k] = a0 + a1*k + a2*k*k + a3*k*k*k;
            errquadsum = (yi[k]-ym[k])*(yi[k]-ym[k]);
        }
    }

    public void ExponentialRegression()
    {
        // Transformation von {x,y} nach {x,ln(y)}
        for (int k=0; k<samples; k++)
        {
            // log(x) -> x>0
            if (yi[k] <= 0) yi[k] = 1; // log(1) = 0

	        yi[k] = Math.Log(yi[k]);
        }

        LinearRegression();

        // Rcktransformation von {x,ln(y)} nach {x,y}
        for (int k=0; k<samples; k++)
        {
            yi[k] = Math.Exp(yi[k]);
        }
        // Rcktransformation des "Achsenabschnitts"
        a0 = Math.Exp(a0);

        // Modelldaten und verbleibende Fehlerquadratsumme berechnen
        for (int k=0; k<samples; k++)
        {
            ym[k] = a0*Math.Exp(a1*k);
            errquadsum = (yi[k]-ym[k])*(yi[k]-ym[k]);
        }
    }

    public void PowerRegression()
    {
        // Transformation von {x,y} nach {ln(x),ln(y)}
        for (int k=0; k<samples; k++)
        {
            // falls k (und damit x[0]) bei 0 beginnt, x[0] = 1 setzen
            if (k == 0) x[0] = 1;

            // log(x) -> x>0
            if (yi[k] <= 0) yi[k] = 1; // log(1) = 0

            x[k]  = Math.Log(x[k]);
            yi[k] = Math.Log(yi[k]);
        }

        LinearRegression();

        // Rcktransformation von {ln(x),ln(y)} nach {x,y}
        for (int k=0; k<samples; k++)
        {
            x[k]  = Math.Exp(x[k]);
            yi[k] = Math.Exp(yi[k]);
        }
        // Rcktransformation des "Achsenabschnitts"
        a0 = Math.Exp(a0);

        // Modelldaten und verbleibende Fehlerquadratsumme berechnen
        for (int k=0; k<samples; k++)
        {
            ym[k] = a0*Math.Pow((double)k, a1);
            errquadsum = (yi[k]-ym[k])*(yi[k]-ym[k]);
        }
    }
}


class Execute
{
    [STAThread]
    static void Main()
    {
        string userinpute = "";
        int select = -1;

        while (select <= 11)
        {
            DataModelling myData = new DataModelling();

            Console.Write("Regressionsanalyse\n\n");

            Console.Write("Bitte Signalform auswhlen:\n\n");
            Console.Write("linear              ->  1\n");
            Console.Write("Polygon 2. Ordnung  ->  2\n");
            Console.Write("Polygon 3. Ordnung  ->  3\n");
            Console.Write("Exponential mit a>0 ->  4\n");
            Console.Write("Exponential mit a<0 ->  5\n");
            Console.Write("Potenz mit a>1      ->  6\n");
            Console.Write("Potenz mit a<1      ->  7\n\n");
            userinpute = Console.ReadLine();
            select = Convert.ToInt32(userinpute);

            myData.CreateSignal(select);

            Console.WriteLine();
            Console.Write("Bitte Funktionstyp fr Regression auswhlen:\n\n");
            Console.Write("linear             ->  1\n");
            Console.Write("Polygon 2. Ordnung ->  2\n");
            Console.Write("Polygon 3. Ordnung ->  3\n");
            Console.Write("Exponential        ->  4\n");
            Console.Write("Potenz             ->  5\n\n");
            userinpute = Console.ReadLine();
            select = Convert.ToInt32(userinpute);

            if (select == 1) myData.LinearRegression();
            if (select == 2) myData.Polynom2Regression();
            if (select == 3) myData.Polynom3Regression();
            if (select == 4) myData.ExponentialRegression();
            if (select == 5) myData.PowerRegression();

            DisplayChart mydisplay = new DisplayChart(myData.x, myData.yi, myData.ym, "Regression", "x(k)", "y(k)");
            mydisplay.Plot2();
            //mydisplay.AddData(0, );
            //mydisplay.AddData(1, myData.x, );


            for (int i = 0; i < 20; i++) Console.WriteLine();
        }

        Console.WriteLine();
        Console.WriteLine("Ende...<ENTER>");
        Console.ReadLine();
    }
}
